from bgtasks.utils.const import RPCStatus


class RPCResponse(dict):
    def __init__(self, data=None, errors=None, **__):
        assert not (data is None and errors is None), \
            'one parameter is acceptable, data or errors'
        if data is not None:
            status = RPCStatus.SUCCESS
            data = data
        else:
            status = RPCStatus.FAIL
            data = errors
        super().__init__(status=status, data=data)



from operator import iconcat
from functools import reduce

try:
    from django.utils.translation import ugettext_lazy as _
    from django.db import models
except ImportError:
    raise ImportError(
        'Please install django framework to use rest part of background task'
    )
try:
    from rest_framework import serializers
    from rest_framework.serializers import ListField
    from rest_framework.serializers import Serializer
    from rest_framework.serializers import IntegerField
    from rest_framework.serializers import LIST_SERIALIZER_KWARGS
except ImportError:
    raise ImportError(
        'Install django rest framework in order to use rest part of background'
        'task'
    )

from .fields import RemoteField
from .response import Response
from ..rpc import RPCClient
from ..utils.const import RPCStatus
from ..utils.merge import merge
from ..constants import FAIL


class RPCListSerializer(serializers.ListSerializer):
    default_error_messages = {
        'required': _('This field is required.'),
        'does_not_exist': _(
            'Invalid pk "{pk_value}" - object does not exist.'),
        'incorrect_type': _(
            'Incorrect type. Expected pk value, received {data_type}.'),
        'timeout': _('Timeout error({route})'),
        'wrong_format': _('Format of response is not correct or route to '
                          'service is incorrect({route})')
    }

    def to_representation(self, data):
        self.child = self.child.__class__(context=dict(many=True))
        iterable = data.all() if isinstance(data, models.Manager) else data
        fields = self.child.fields.fields
        rpc_client = RPCClient()

        rpc_fields = dict()
        for f, v in fields.items():
            if isinstance(v, RemoteField):
                rpc_fields[f] = dict(
                    route=v.route,
                    source=v.source,
                    key=v.key,
                    merge_key=v.merge_key
                )
        for field, data in rpc_fields.items():
            raw_values = list()
            for item in iterable:
                internal_type = item._meta.get_field(
                    data['source']
                ).get_internal_type()
                is_array = internal_type == 'ArrayField'

                if isinstance(item, dict):
                    raw_values.append(item.get(data['source']))
                else:
                    raw_values.append(getattr(item, data['source'], None))
            raw_values = list(filter(None, raw_values))
            data['raw_values'] = raw_values
            if not raw_values:
                continue
            try:
                if is_array:
                    array_values = raw_values
                    raw_values = reduce(iconcat, raw_values, [])
                rpc_response = rpc_client.call(
                    data['route'], {data['key']: raw_values})
                if not RPCStatus.is_success(rpc_response):
                    raise serializers.ValidationError(rpc_response['data'])
                if is_array:
                    rpc_response_data = rpc_response['data']
                    data['obj_values'] = []
                    for value in array_values:
                        data['obj_values'].append(
                            rpc_response_data[:len(value)]
                        )
                        del rpc_response_data[:len(value)]
                else:
                    data['obj_values'] = rpc_response['data']
            except TimeoutError:
                self.fail('timeout', route=data['route'])
            except KeyError:
                self.fail('wrong_format', route=data['route'])
        for key, value in self.context.items():
            setattr(self.child, key, value)
        response_data = [
            self.child.to_representation(item) for item in iterable
        ]

        for field, data in rpc_fields.items():
            if not data.get('obj_values'):
                continue
            for r in response_data:
                merge(r, field, data['obj_values'], data['merge_key'])
        return response_data


class RPCSerializerMixin(object):
    @classmethod
    def many_init(cls, *args, **kwargs):
        allow_empty = kwargs.pop('allow_empty', None)
        child_serializer = cls(*args, **kwargs)
        list_kwargs = {
            'child': child_serializer,
        }
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty
        list_kwargs.update({
            key: value for key, value in kwargs.items()
            if key in LIST_SERIALIZER_KWARGS
        })
        meta = getattr(cls, 'Meta', None)
        list_serializer_class = getattr(
            meta, 'list_serializer_class', RPCListSerializer)
        return list_serializer_class(*args, **list_kwargs)


class IdsSerializer(serializers.Serializer):
    ids = ListField(child=IntegerField(), required=False, allow_empty=False)

    def update(self, instance, validated_data):
        raise NotImplemented

    def create(self, validated_data):
        raise NotImplemented


def serializer_class(klass=IdsSerializer, **kwargs):
    def wrapper(func):
        def inner(data):
            serializer = klass(data=data, **kwargs)

            if serializer.is_valid():
                return func(serializer)
            return Response(serializer.errors, FAIL)
        return inner

    return wrapper



from bgtasks import RPCClient

try:
    from django.utils.translation import ugettext_lazy as _
except ImportError:
    raise ImportError(
        'Please install django framework to use rest part of background task'
    )
try:
    from rest_framework import serializers
except ImportError:
    raise ImportError(
        'Install django rest framework in order to use rest part of background'
        'task'
    )


class RemoteField(serializers.RelatedField):
    default_error_messages = {
        'required': _('This field is required.'),
        'does_not_exist': _('Invalid pk "{pk_value}" '
                            '- object does not exist.'),
        'incorrect_type': _('Incorrect type. Expected pk value,'
                            ' received {data_type}.'),
        'timeout': _('Timeout error({route})'),
        'wrong_format': _('Format of response is not correct or route to '
                          'service is incorrect({route})')
    }

    def __init__(self, route, key='ids', merge_key='id', **kwargs):
        self.route = route
        self.key = key
        self.merge_key = merge_key
        self.rpc_client = RPCClient()
        self.response_data = dict()
        if not kwargs.get('read_only', False):
            kwargs.update(queryset=[])
        self.property_field: str = kwargs.pop('property_field', None)
        super(RemoteField, self).__init__(**kwargs)

    def rpc_call(self, body):
        response = self.rpc_client.call(self.route, body)
        status = response['status']
        data = response['data']

        return data, status

    def to_internal_value(self, value):
        body = {self.key: [value]}

        try:
            self.response_data, status = self.rpc_call(body)

            if status == 'fail':
                raise serializers.ValidationError(self.response_data)

            if not len(self.response_data):
                raise self.fail('does_not_exist', pk_value=value)

            return value
        except TimeoutError:
            self.fail('timeout', route=self.route)

        except KeyError:
            self.fail('wrong_format', route=self.route)

    def to_representation(self, value):
        if self.context.get('many', False):
            return value
        if not self.response_data:
            self.to_internal_value(value)
        return self.response_data[0]



from django.db import connection


class TwoPhaseCommit(object):
    cursor = None

    def __enter__(self):
        self.cursor = connection.cursor()
        xid = self.cursor.db.connection.xid()
        self.cursor.db.set_autocommit(False)
        self.cursor.db.connection.tpc_begin(xid)

        return str(xid)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cursor.db.connection.tpc_prepare()



import unittest


class UtilTest(unittest.TestCase):
    def test_response(self):
        from bgtasks.utils.const import RPCStatus
        from bgtasks.utils.rpc_response import RPCResponse

        response = RPCResponse(data=dict(response='successful response'))
        self.assertEqual(RPCStatus.is_success(response), True)

        response = RPCResponse(errors=dict(response='successful response'))
        self.assertEqual(RPCStatus.is_success(response), False)

    def test_merge_dict(self):
        from bgtasks.utils.merge import merge_dict

        data = [
            dict(product=1, quantity=20),
            dict(product=5, quantity=10),
        ]
        merge_data = [
            dict(id=1, name='Product 1', code='0003'),
            dict(id=2, name='Product 2', code='0001'),
            dict(id=3, name='Product 3', code='0002'),
            dict(id=5, name='Product 5', code='0002'),
        ]

        for datum in data:
            merge_dict(datum, 'product', merge_data, 'id')
            self.assertEqual(isinstance(datum['product'], dict), True)

        data = [
            dict(product=10, quantity=20),
            dict(product=4, quantity=10),
            dict(product=15, quantity=10),
        ]

        for datum in data:
            merge_dict(datum, 'product', merge_data, 'id')
            self.assertEqual(isinstance(datum['product'], dict), False)

    def test_merge_obj(self):
        from bgtasks.utils.merge import merge_obj

        class Foo(object):
            def __init__(self, **kwargs):
                for key, value in kwargs.items():
                    setattr(self, key, value)

        data = [
            Foo(product=1, quantity=20),
            Foo(product=5, quantity=10),
        ]
        merge_data = [
            dict(id=1, name='Product 1', code='0003'),
            dict(id=2, name='Product 2', code='0001'),
            dict(id=3, name='Product 3', code='0002'),
            dict(id=5, name='Product 5', code='0002'),
        ]

        for datum in data:
            merge_obj(datum, 'product', merge_data, 'id')
            self.assertEqual(isinstance(datum.product, dict), True)

        data = [
            Foo(product=20, quantity=20),
            Foo(product=4, quantity=10),
            Foo(product=15, quantity=10),
        ]

        for datum in data:
            merge_obj(datum, 'product', merge_data, 'id')
            self.assertEqual(isinstance(datum.product, dict), False)



from ..constants import SUCCESS, FAIL


class IncorrectStatus(Exception):
    pass


class Response(dict):
    def __init__(self, data, status=SUCCESS):
        if status != FAIL and status != SUCCESS:
            raise IncorrectStatus('Status should be equal fail or success')

        super().__init__(status=status, data=data)



from .rpc import rpc_tasks, RPCClient
from .job import job_tasks, JobClient
from .rest_framework.serilaizers import serializer_class
from .rest_framework.response import Response
from .rest_framework.fields import RemoteField
from .constants import SUCCESS, FAIL

name = "bgtasks"

__all__ = [
    'name', 'rpc_tasks', 'job_tasks', 'serializer_class',
    'RemoteField', 'Response', 'SUCCESS', 'FAIL',
    'RPCClient', 'JobClient'
]



from django.db import models


class RemoteField(models.PositiveIntegerField):
    def __init__(self, **kwargs):
        kwargs.setdefault('db_index', True)
        super().__init__(**kwargs)

    def get_column_name(self):
        return '%s_id' % self.name

    def get_attname_column(self):
        attname = self.get_attname()
        column = self.get_column_name()
        return attname, column



# Generated by Django 2.2.6 on 2019-10-02 07:09

import bgtasks.models
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='App',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('app', bgtasks.models.RemoteField(db_index=True)),
            ],
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('app', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='testapp.App')),
            ],
        ),
    ]



import django_filters


class DefaultFilterSet(django_filters.FilterSet):
    ids = django_filters.NumberFilter(method='filter_by_ids')

    def filter_by_ids(self, queryset, name, value):
        return queryset.filter(id__in=value)



from django.test import TestCase
from bgtasks import serializer_class
from ..rpc import rpc_tasks, RPCClient
from ..constants import SUCCESS, FAIL
from ..rest_framework.response import Response


@rpc_tasks('task1')
def task1(_):
    return Response('Hi 1')


@rpc_tasks('task2')
@serializer_class()
def task2(_):
    return Response('Hi 2')


class RPCTasksTest(TestCase):
    def test_task1(self):
        response = RPCClient().call('task1', {})
        self.assertEqual(SUCCESS, response['status'])
        self.assertEqual('Hi 1', response['data'])

    def test_serializer_class_1(self):
        message = 'Expected a list of items but got type "int".'
        response = RPCClient().call('task2', {'ids': 1})
        self.assertEqual(FAIL, response['status'])
        self.assertEqual(message, response['data']['ids'][0])

    def test_serializer_class_2(self):
        message = 'Expected a list of items but got type "str".'
        response = RPCClient().call('task2', {'ids': 'hello'})
        self.assertEqual(FAIL, response['status'])
        self.assertEqual(message, response['data']['ids'][0])

    def test_serializer_class_3(self):
        message = 'This list may not be empty.'
        response = RPCClient().call('task2', {'ids': []})
        self.assertEqual(FAIL, response['status'])
        self.assertEqual(message, response['data']['ids'][0])

    def test_serializer_class_4(self):
        message = 'No data provided'
        response = RPCClient().call('task2', None)
        self.assertEqual(FAIL, response['status'])
        self.assertEqual(message, response['data']['non_field_errors'][0])

    def test_serializer_class_5(self):
        response = RPCClient().call('task2', {'ids': [1, 2, 3]})
        self.assertEqual(SUCCESS, response['status'])



from django.db import models
from bgtasks.models import RemoteField


# Create your models here.
class App(models.Model):
    app = RemoteField()


class User(models.Model):
    app = models.ForeignKey(App, on_delete=models.PROTECT)


class Category(models.Model):
    name = models.CharField(max_length=255)



from ..constants import SUCCESS, FAIL


class RPCStatus:
    FAIL = FAIL
    SUCCESS = SUCCESS

    @classmethod
    def is_success(cls, data):
        if isinstance(data, str):
            return data == cls.SUCCESS
        return data.get('status', cls.FAIL) == cls.SUCCESS



def merge(obj, field, merge_dicts: list, merge_field, raise_exception=False):
    for m in merge_dicts:
        if type(m) in [list, tuple]:
            merge_value = [i[merge_field] for i in m]
        else:
            merge_value = m.get(merge_field)

        if isinstance(obj, dict):
            value = obj.get(field)
            if value == merge_value:
                obj[field] = m
                return obj
        if isinstance(obj, object):
            value = getattr(obj, field, None)
            if value == merge_value:
                setattr(obj, field, m[merge_field])
                return obj
    if raise_exception:
        raise ValueError('Can not merge as data is not found')
    return


def merge_dict(obj: dict, field, merge_dicts: list,
               merge_field, raise_exception=False):
    value = obj.get(field, None)
    for m in merge_dicts:
        if value == m.get(merge_field):
            obj[field] = m
            return obj
    if raise_exception:
        raise ValueError('Can not merge as data is not found')
    return


def merge_obj(obj: object, field, merge_dicts: list,
              merge_field, raise_exception=False):
    value = getattr(obj, field, None)
    for m in merge_dicts:
        if value == m.get(merge_field):
            setattr(obj, field, m)
            return obj
    if raise_exception:
        raise ValueError('Can not merge as data is not found')
    return
